import unittest
import colorama
import statusbar


class TestProgressBar(unittest.TestCase):
    """Test of the progressbar part of a status bar."""

    def test_chunk_widths(self):
        pb = statusbar.ProgressBar(sep_start="", sep_end="")
        pb.add_progress(1, '.')
        pb.add_progress(1, '#')
        breakpoints = pb._get_chunk_sizes(2)
        self.assertListEqual(breakpoints, [1, 1])

        pb = statusbar.ProgressBar(sep_start="", sep_end="")
        pb.add_progress(1, '.')
        pb.add_progress(2, '#')
        breakpoints = pb._get_chunk_sizes(3)
        self.assertListEqual(breakpoints, [1, 2])

        # squeezing [1,2] into width 2 will give 1 to each since
        # the first breakpoint is at 2/3 which is rounded up.
        pb = statusbar.ProgressBar(sep_start="", sep_end="")
        pb.add_progress(1, '.')
        pb.add_progress(2, '#')
        breakpoints = pb._get_chunk_sizes(2)
        self.assertListEqual(breakpoints, [1, 1])

        # squeezing [1,3] into width 2 will give (0,1).
        # The first breakpoint is at 1/2 which *should* be rounded up
        # but actually round(0.5) seems to be 0 in Python...
        pb = statusbar.ProgressBar(sep_start="", sep_end="")
        pb.add_progress(1, '.')
        pb.add_progress(3, '#')
        breakpoints = pb._get_chunk_sizes(2)
        self.assertListEqual(breakpoints, [0, 2])

        pb = statusbar.ProgressBar(sep_start="", sep_end="")
        pb.add_progress(1, '.')
        pb.add_progress(3, '#')
        breakpoints = pb._get_chunk_sizes(4)
        self.assertListEqual(breakpoints, [1, 3])

    def test_progress_formatting(self):
        pb = statusbar.ProgressBar()
        pb.add_progress(1, '.')
        pb.add_progress(1, '#')
        progress = pb.format_progress(4)
        self.assertEqual(progress, "[.#]")

        pb = statusbar.ProgressBar()
        pb.add_progress(1, '.')
        pb.add_progress(2, '#')
        progress = pb.format_progress(5)
        self.assertEqual(progress, "[.##]")

        # Adding a forground colour makes each segment ten characters
        # longer; five characters are used for setting the color and another
        # five for resetting it again. These are not shown, so the width
        # doesn't take this into account.
        pb = statusbar.ProgressBar()
        pb.add_progress(1, '.', fg=colorama.Fore.GREEN)
        pb.add_progress(2, '#', fg=colorama.Fore.RED)
        progress = pb.format_progress(5)
        self.assertEqual(progress[0], "[")
        self.assertEqual(progress[1:6], colorama.Fore.GREEN)
        self.assertEqual(progress[6], ".")
        self.assertEqual(progress[7:12], colorama.Fore.RESET)
        self.assertEqual(progress[12:17], colorama.Fore.RED)
        self.assertEqual(progress[17:19], "##")
        self.assertEqual(progress[19:24], colorama.Fore.RESET)
        self.assertEqual(progress[24], "]")

        # Adding a background and style as well makes the string even longer
        # but other than that there is nothing surprising going on.
        pb = statusbar.ProgressBar()
        pb.add_progress(1, '.',
                        fg=colorama.Fore.GREEN, bg=colorama.Back.RED,
                        style=colorama.Style.BRIGHT)
        pb.add_progress(2, '#',
                        fg=colorama.Fore.RED, bg=colorama.Back.GREEN,
                        style=colorama.Style.DIM)
        progress = pb.format_progress(5)
        self.assertEqual(progress[0], "[")

        self.assertEqual(progress[1:5], colorama.Style.BRIGHT)
        self.assertEqual(progress[5:10], colorama.Back.RED)
        self.assertEqual(progress[10:15], colorama.Fore.GREEN)
        self.assertEqual(progress[15], ".")
        self.assertEqual(progress[16:21], colorama.Fore.RESET)
        self.assertEqual(progress[21:26], colorama.Back.RESET)
        self.assertEqual(progress[26:30], colorama.Style.RESET_ALL)

        self.assertEqual(progress[30:34], colorama.Style.DIM)
        self.assertEqual(progress[34:39], colorama.Back.GREEN)
        self.assertEqual(progress[39:44], colorama.Fore.RED)
        self.assertEqual(progress[44:46], "##")
        self.assertEqual(progress[46:51], colorama.Fore.RESET)
        self.assertEqual(progress[51:56], colorama.Back.RESET)
        self.assertEqual(progress[56:60], colorama.Style.RESET_ALL)

        self.assertEqual(progress[60], "]")

    def test_summary_string(self):
        pb = statusbar.ProgressBar()
        pb.add_progress(1, '.')
        pb.add_progress(1, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 3)
        self.assertEqual(len(summary_string), 3)
        self.assertEqual(summary_string, "1/1")

        pb = statusbar.ProgressBar()
        pb.add_progress(0, '.')
        pb.add_progress(1, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 3)
        self.assertEqual(len(summary_string), 3)
        self.assertEqual(summary_string, "0/1")

        pb = statusbar.ProgressBar()
        pb.add_progress(0, '.')
        pb.add_progress(0, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 3)
        self.assertEqual(len(summary_string), 3)
        self.assertEqual(summary_string, "0/0")

        pb = statusbar.ProgressBar()
        pb.add_progress(9, '.')
        pb.add_progress(9, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 3)
        self.assertEqual(len(summary_string), 3)
        self.assertEqual(summary_string, "9/9")

        pb = statusbar.ProgressBar()
        pb.add_progress(10, '.')
        pb.add_progress(10, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 5)
        self.assertEqual(len(summary_string), 5)
        self.assertEqual(summary_string, "10/10")

        pb = statusbar.ProgressBar()
        pb.add_progress(99, '.')
        pb.add_progress(99, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 5)
        self.assertEqual(len(summary_string), 5)
        self.assertEqual(summary_string, "99/99")

        pb = statusbar.ProgressBar()
        pb.add_progress(100, '.')
        pb.add_progress(199, '#')
        estimated_length = pb.summary_width()
        summary_string = pb.format_summary()
        self.assertEqual(estimated_length, 7)
        self.assertEqual(len(summary_string), 7)
        self.assertEqual(summary_string, "100/199")


class TestStatusBar(unittest.TestCase):
    """Test of a status bar."""

    def test_status_formatting(self):
        sb = statusbar.StatusBar("Test")
        sb.add_progress(2, '#')
        sb.add_progress(2, '.')
        result = sb.format_status(15)
        self.assertEqual(result[:5], "Test ")
        self.assertEqual(result[-4:], " 2/2")
        self.assertEqual(result[5:-4], "[##..]")

        sb.set_progress_brackets("", "")
        result = sb.format_status(15)
        self.assertEqual(result[:5], "Test ")
        self.assertEqual(result[-4:], " 2/2")
        self.assertEqual(result[5:-4], "###...")

        sb = statusbar.StatusBar("Long label")
        sb.add_progress(2, '#')
        sb.add_progress(2, '.')
        result = sb.format_status(15, label_width=4)
        self.assertEqual(result[:5], "L... ")
        self.assertEqual(result[-4:], " 2/2")
        self.assertEqual(result[5:-4], "[##..]")

        sb = statusbar.StatusBar("Long label")
        sb.add_progress(2, '#')
        sb.add_progress(2, '.')
        result = sb.format_status(26, label_width=15)
        self.assertEqual(result[:16], "Long label..... ")
        self.assertEqual(result[-4:], " 2/2")
        self.assertEqual(result[16:-4], "[##..]")

        result = sb.format_status(label_width=4,
                                  progress_width=10)
        self.assertEqual(result[:5], "L... ")
        self.assertEqual(result[-4:], " 2/2")
        self.assertEqual(result[5:-4], "[####....]")

        result = sb.format_status(label_width=4,
                                  progress_width=10,
                                  summary_width=5)
        self.assertEqual(result[:5], "L... ")
        self.assertEqual(result[-6:], "   2/2")
        self.assertEqual(result[5:-6], "[####....]")


class TestStatusTable(unittest.TestCase):
    """Test of a status table."""

    def test_field_width_calculations(self):
        st = statusbar.StatusTable()

        label1 = "Test"
        sb = st.add_status_line(label1)
        sb.add_progress(1, "#")
        sb.add_progress(1, " ")

        label2 = "Testing progress"
        sb = st.add_status_line(label2)
        sb.add_progress(10, "#")
        sb.add_progress(10, " ")

        self.assertEqual(st.label_width(), len(label2))
        self.assertEqual(st.summary_width(), len("10/10"))

        labw, progw, sumw = st.calculate_field_widths(width=30)
        self.assertEqual(sumw, 5)
        self.assertEqual(progw, 10)
        self.assertEqual(labw, 15 - 2)

        labw, progw, sumw = st.calculate_field_widths(width=40)
        self.assertEqual(sumw, 5)
        self.assertEqual(labw, len(label2))
        self.assertEqual(progw, 40 - len(label2) - 5 - 2)

        labw, progw, sumw = st.calculate_field_widths(width=1)
        self.assertEqual(sumw, 5)
        self.assertEqual(progw, 10)
        self.assertEqual(labw, 10)

    def test_table_formatting(self):
        st = statusbar.StatusTable()

        label1 = "Test"
        sb = st.add_status_line(label1)
        sb.add_progress(1, "#")
        sb.add_progress(1, " ")

        label2 = "Testing progress"
        sb = st.add_status_line(label2)
        sb.add_progress(10, "#")
        sb.add_progress(10, " ")

        formatted = st.format_table(width=40)
        labelw = len(label2)
        statusw = len("10/10")

        first_label = formatted[0][:labelw]
        second_label = formatted[1][:labelw]
        first_status = formatted[0][-statusw:]
        second_status = formatted[1][-statusw:]

        self.assertEqual(first_label, "Test............")
        self.assertEqual(second_label, label2)
        self.assertEqual(first_status, "  1/1")
        self.assertEqual(second_status, "10/10")



import os
from setuptools import find_packages, setup


# Utility function to read the README file.  Used for the
# long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put
# a raw string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name="statusbar",
    version="0.1.1",

    packages=find_packages(),

    test_suite='tests',
    install_requires=[
        "colorama>=0.3.7",
    ],

    # metadata for upload to PyPI
    author="Thomas Mailund",
    author_email="mailund@birc.au.dk",
    license="GPLv3",
    keywords="CLI status output",
    url="https://github.com/mailund/statusbar",
    description="Module for displaying (text based) status update lines.",
    long_description=read('README.rst'),

    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",    # NOQA
        "Programming Language :: Python",
    ],
)



"""Module for constructing text-based status updates."""

import shutil
import itertools
import colorama
from math import log10, ceil

colorama.init()


class _ProgressChunk:
    def __init__(self, count, symbol, fg, bg, style):
        self.count = count
        self.symbol = symbol
        self.fg = fg
        self.bg = bg
        self.style = style

        if self.fg is None:
            self.fg = ''
            self.fg_reset = ''
        else:
            self.fg_reset = colorama.Fore.RESET

        if self.bg is None:
            self.bg = ''
            self.bg_reset = ''
        else:
            self.bg_reset = colorama.Back.RESET

        if self.style is None:
            self.style = ''
            self.style_reset = ''
        else:
            self.style_reset = colorama.Style.RESET_ALL

    def format_chunk(self, width):
        return "{format_start}{bar}{format_end}".format(
            format_start=self.style + self.bg + self.fg,
            bar=self.symbol * width,
            format_end=self.fg_reset + self.bg_reset + self.style_reset
        )

    def format_chunk_summary(self):
        return "{format_start}{count}{format_end}".format(
            format_start=self.style + self.bg + self.fg,
            count=self.count,
            format_end=self.fg_reset + self.bg_reset + self.style_reset
        )


class ProgressBar:
    """Class responsible for showing progress of a task."""

    def __init__(self, sep_start='[', sep_end=']'):
        """Construct a progress bar."""
        self._progress_chunks = []
        self.sep_start = sep_start
        self.sep_end = sep_end

    def set_progress_brackets(self, start, end):
        """Set brackets to set around a progress bar."""
        self.sep_start = start
        self.sep_end = end

    def add_progress(self, count, symbol='#',
                     fg=None, bg=None, style=None):
        """Add a section of progress to the progressbar.

        The progress is captured by "count" and displayed as a fraction
        of the statusbar width proportional to this count over the total
        progress displayed. The progress will be displayed using the "symbol"
        character and the foreground and background colours and display style
        determined by the the "fg", "bg" and "style" parameters. For these,
        use the colorama package to set up the formatting.
        """
        chunk = _ProgressChunk(count, symbol, fg, bg, style)
        self._progress_chunks.append(chunk)

    def _get_chunk_sizes(self, width):

        # FIXME: using len() here doesn't work with invisible symbols
        # or things like tabs that take up one character but uses more
        # space in the terminal.
        width = width - len(self.sep_start + self.sep_end)

        chunk_counts = [chunk.count for chunk in self._progress_chunks]
        total = sum(chunk_counts)
        chunk_real_widths = [chunk/total*width for chunk in chunk_counts]
        chunk_break_points = \
            [int(round(bp)) for bp in itertools.accumulate(chunk_real_widths)]

        start_points = [0] + chunk_break_points[:-1]
        end_points = chunk_break_points
        chunk_widths = \
            [(end-start) for (start, end) in zip(start_points, end_points)]

        assert sum(chunk_widths) == width
        return chunk_widths

    def format_progress(self, width):
        """Create the formatted string that displays the progress."""
        chunk_widths = self._get_chunk_sizes(width)
        progress_chunks = [chunk.format_chunk(chunk_width)
                           for (chunk, chunk_width)
                           in zip(self._progress_chunks, chunk_widths)]
        return "{sep_start}{progress}{sep_end}".format(
            sep_start=self.sep_start,
            progress="".join(progress_chunks),
            sep_end=self.sep_end
        )

    def summary_width(self):
        """Calculate how long a string is needed to show a summary string.

        This is not simply the length of the formatted summary string
        since that string might contain ANSI codes.
        """
        chunk_counts = [chunk.count for chunk in self._progress_chunks]
        numbers_width = sum(max(1, ceil(log10(count + 1)))
                            for count in chunk_counts)
        separators_with = len(chunk_counts) - 1
        return numbers_width + separators_with

    def format_summary(self):
        """Generate a summary string for the progress bar."""
        chunks = [chunk.format_chunk_summary()
                  for chunk in self._progress_chunks]
        return "/".join(chunks)


class StatusBar:
    """Class for displaying status bars.

    These bars consists of a label, a progress bar, and statistics/summaries
    of the progress.
    """

    def __init__(self, label,
                 progress_sep_start='[', progress_sep_end=']'):
        """Construct a status bar."""
        self.label = label
        self._progress = ProgressBar(progress_sep_start, progress_sep_end)

    def set_progress_brackets(self, start, end):
        """Define which braces should be used around the progress bar."""
        self._progress.set_progress_brackets(start, end)

    def add_progress(self, count, symbol='#',
                     fg=None, bg=None, style=None):
        """Add a section of progress to the progressbar.

        The progress is captured by "count" and displayed as a fraction
        of the statusbar width proportional to this count over the total
        progress displayed. The progress will be displayed using the "symbol"
        character and the foreground and background colours and display style
        determined by the the "fg", "bg" and "style" parameters. For these,
        use the colorama package to set up the formatting.
        """
        self._progress.add_progress(count, symbol, fg, bg, style)

    def summary_width(self):
        """Get the minimum width the progress summary field will use."""
        return self._progress.summary_width()

    def label_width(self):
        """Get the minimum width the progress label field will use."""
        return len(self.label)

    def format_status(self, width=None,
                      label_width=None,
                      progress_width=None,
                      summary_width=None):
        """Generate the formatted status bar string."""
        if width is None:  # pragma: no cover
            width = shutil.get_terminal_size()[0]

        if label_width is None:
            label_width = len(self.label)
        if summary_width is None:
            summary_width = self.summary_width()
        if progress_width is None:
            progress_width = width - label_width - summary_width - 2

        if len(self.label) > label_width:
            # FIXME: This actually *will* break if we ever have fewer than
            # three characters assigned to format the label, but that would
            # be an extreme situation so I won't fix it just yet.
            label = self.label[:label_width - 3] + "..."
        else:
            label_format = "{{label:.<{width}}}".format(width=label_width)
            label = label_format.format(label=self.label)

        summary_format = "{{:>{width}}}".format(width=summary_width)
        summary = summary_format.format(self._progress.format_summary())

        progress = self._progress.format_progress(width=progress_width)

        return "{label} {progress} {summary}".format(
            label=label,
            progress=progress,
            summary=summary
        )


class StatusTable:
    """Several lines of status bars with the three fields aligned."""

    def __init__(self, progress_sep_start='[', progress_sep_end=']'):
        """Create a status table."""
        self._lines = []
        self._sep_start = progress_sep_start
        self._sep_end = progress_sep_end

    def add_status_line(self, label):
        """Add a status bar line to the table.

        This function returns the status bar and it can be modified
        from this return value.
        """
        status_line = StatusBar(label, self._sep_start, self._sep_end)
        self._lines.append(status_line)
        return status_line

    def summary_width(self):
        """Compute the minimum size needed for the summary field."""
        return max(
            sb.summary_width() for sb in self._lines
        )

    def label_width(self):
        """Compute the minimum size needed for the label field.

        This is the minimum size needed if all labels are shown in full.
        If there is not room on a line to shown them in full they will
        be truncated."""
        return max(
            sb.label_width() for sb in self._lines
        )

    def calculate_field_widths(self, width=None,
                               min_label_width=10,
                               min_progress_width=10):
        """Calculate how wide each field should be so we can align them.

        We always find room for the summaries since these are short and
        packed with information. If possible, we will also find room for
        labels, but if this would make the progress bar width shorter than
        the specified minium then we will shorten the labels, though never
        below the minium there. If this mean we have bars that are too wide
        for the terminal, then your terminal needs to be wider.
        """
        if width is None:  # pragma: no cover
            width = shutil.get_terminal_size()[0]

        summary_width = self.summary_width()
        label_width = self.label_width()
        remaining = width - summary_width - label_width - 2

        if remaining >= min_progress_width:
            progress_width = remaining
        else:
            progress_width = min_progress_width
            remaining = width - summary_width - progress_width - 2
            if remaining >= min_label_width:
                label_width = remaining
            else:
                label_width = min_label_width

        return (label_width, progress_width, summary_width)

    def format_table(self, width=None,
                     min_label_width=10, min_progress_width=10):
        """Format the entire table of progress bars.

        The function first computes the widths of the fields so they can be
        aligned across lines and then returns formatted lines as a list of
        strings.
        """
        if width is None:  # pragma: no cover
            width = shutil.get_terminal_size()[0]

        labelw, progw, summaryw = self.calculate_field_widths(
            width=width,
            min_label_width=min_label_width,
            min_progress_width=min_progress_width
        )
        output = [
            sb.format_status(
                label_width=labelw,
                progress_width=progw,
                summary_width=summaryw
            )
            for sb in self._lines
        ]

        return output
